@@ -935,7 +935,7 @@ class BufferContext {
               readonly start: number) {}
 }
 
-class BufferNode extends BaseNode {
+export class BufferNode extends BaseNode {
   type: NodeType
 
   get name() { return this.type.name }
@@ -1013,8 +1013,6 @@ class BufferNode extends BaseNode {
 
 function iterStack(heads: readonly SyntaxNode[]): NodeIterator | null {
   if (!heads.length) return null
-  if (heads.length == 1) return heads[0] as any
-
   let pick = 0, picked = heads[0]
   for (let i = 1; i < heads.length; i++) {
     let node = heads[i]
@@ -1045,7 +1043,7 @@ function stackIterator(tree: Tree, pos: number, side: -1 | 0 | 1): NodeIterator
       let mount = MountedTree.get(scan.tree)
       // Relevant overlay branching off
       if (mount && mount.overlay && mount.overlay[0].from <= pos && mount.overlay[mount.overlay.length - 1].to >= pos) {
-        let root = new TreeNode(mount.tree, mount.overlay[0].from + scan.from, 0, null)
+        let root = new TreeNode(mount.tree, mount.overlay[0].from + scan.from, -1, scan)
         ;(layers || (layers = [inner])).push(resolveNode(root, pos, side, false))
       }
     }
@@ -1112,7 +1110,8 @@ export class TreeCursor implements SyntaxNodeRef {
     return true
   }
 
-  private yield(node: TreeNode | BufferNode | null) {
+  /// @internal
+  yield(node: TreeNode | BufferNode | null) {
     if (!node) return false
     if (node instanceof TreeNode) {
       this.buffer = null
@@ -1333,6 +1332,7 @@ function hasChild(tree: Tree): boolean {
 }
 
 const enum Balance { BranchFactor = 8 }
+const enum CutOff { Depth = 2500 }
 
 const enum SpecialRecord {
   Reuse = -1,
@@ -1352,7 +1352,7 @@ function buildTree(data: BuildData) {
 
   function takeNode(parentStart: number, minPos: number,
                     children: (Tree | TreeBuffer)[], positions: number[],
-                    inRepeat: number) {
+                    inRepeat: number, depth: number) {
     let {id, start, end, size} = cursor
     let lookAheadAtStart = lookAhead
     while (size < 0) {
@@ -1393,13 +1393,16 @@ function buildTree(data: BuildData) {
       while (cursor.pos > endPos) {
         if (localInRepeat >= 0 && cursor.id == localInRepeat && cursor.size >= 0) {
           if (cursor.end <= lastEnd - maxBufferLength) {
-            makeRepeatLeaf(localChildren, localPositions, start, lastGroup, cursor.end, lastEnd, localInRepeat, lookAheadAtStart)
+            makeRepeatLeaf(localChildren, localPositions, start, lastGroup, cursor.end, lastEnd,
+                           localInRepeat, lookAheadAtStart)
             lastGroup = localChildren.length
             lastEnd = cursor.end
           }
           cursor.next()
+        } else if (depth > CutOff.Depth) {
+          takeFlatNode(start, endPos, localChildren, localPositions)
         } else {
-          takeNode(start, endPos, localChildren, localPositions, localInRepeat)
+          takeNode(start, endPos, localChildren, localPositions, localInRepeat, depth + 1)
         }
       }
       if (localInRepeat >= 0 && lastGroup > 0 && lastGroup < localChildren.length)
@@ -1418,6 +1421,37 @@ function buildTree(data: BuildData) {
     positions.push(startPos)
   }
 
+  function takeFlatNode(parentStart: number, minPos: number,
+                        children: (Tree | TreeBuffer)[], positions: number[]) {
+    let nodes = [] // Temporary, inverted array of leaf nodes found, with absolute positions
+    let nodeCount = 0, stopAt = -1
+    while (cursor.pos > minPos) {
+      let {id, start, end, size} = cursor
+      if (size > 4) { // Not a leaf
+        cursor.next()
+      } else if (stopAt > -1 && start < stopAt) {
+        break
+      } else {
+        if (stopAt < 0) stopAt = end - maxBufferLength
+        nodes.push(id, start, end)
+        nodeCount++
+        cursor.next()
+      }
+    }
+    if (nodeCount) {
+      let buffer = new Uint16Array(nodeCount * 4)
+      let start = nodes[nodes.length - 2]
+      for (let i = nodes.length - 3, j = 0; i >= 0; i -= 3) {
+        buffer[j++] = nodes[i]
+        buffer[j++] = nodes[i + 1] - start
+        buffer[j++] = nodes[i + 2] - start
+        buffer[j++] = j
+      }
+      children.push(new TreeBuffer(buffer, nodes[2] - start, nodeSet))
+      positions.push(start - parentStart)
+    }
+  }
+
   function makeBalanced(type: NodeType) {
     return (children: readonly (Tree | TreeBuffer)[], positions: readonly number[], length: number) => {
       let lookAhead = 0, lastI = children.length - 1, last, lookAheadProp
@@ -1522,7 +1556,7 @@ function buildTree(data: BuildData) {
   }
 
   let children: (Tree | TreeBuffer)[] = [], positions: number[] = []
-  while (cursor.pos > 0) takeNode(data.start || 0, data.bufferStart || 0, children, positions, -1)
+  while (cursor.pos > 0) takeNode(data.start || 0, data.bufferStart || 0, children, positions, -1, 0)
   let length = data.length ?? (children.length ? positions[0] + children[0].length : 0)
   return new Tree(types[data.topID], children.reverse(), positions.reverse(), length)
 }
